from __future__ import annotations
import argparse
import textwrap
from functools import partial
from typing import Callable, List, Tuple, Dict

from educelab.cmdparse import parse, parse_parameter_list
from numpy.typing import ArrayLike

from .adjust import brightness_contrast, exposure, shadows
from .correction import gamma_correction, normalize
from .enhance import clahe, clip, curves, stretch, stretch_percentile
from .filters import sharpen

CommandStrList = List[str]
Command = Tuple[str, Dict]
CommandList = List[Command]
Pipeline = Callable[[ArrayLike], ArrayLike]

_fn_map = {
    'brightness-contrast': brightness_contrast,
    'clahe': clahe,
    'clip': clip,
    'curves': curves,
    'exposure': exposure,
    'gamma': gamma_correction,
    'normalize': normalize,
    'shadows': shadows,
    'sharpen': sharpen,
    'stretch': stretch,
    'pstretch': stretch_percentile
}


def commands_description() -> str:
    return textwrap.dedent("""\
    -brightness-contrast=BRIGHTNESS{,CONTRAST}
    \t\t\tBrightness/contrast adjustment
    -clahe{=KERNEL{,BINS}}
    \t\t\tContrast Limited Adaptive Histogram Equalization
    -clip{=MIN{,MAX}}
    \t\t\tClip values to range
    -curves
    \t\t\tCurves enhancement. Currently a preset enhancement curve.
    -exposure{=VAL}
    \t\t\tAdjust image exposure
    -gamma{=GAMMA{,GAIN}}
    \t\t\tGamma correction
    -normalize
    \t\t\tLinear contrast stretch to data min/max
    -shadows=VAL
    \t\t\tAdjust shadow brightness
    -sharpen{=RADIUS{,AMOUNT}}
    \t\t\tUnsharp masking filter
    -stretch=MIN,MAX
    \t\t\tLinear contrast stretch to absolute values
    -pstretch=MIN,MAX
    \t\t\tLinear contrast stretch to data percentiles
    """)


class Parsers:
    """cmdparse namespace for parsing enhancement command arguments."""

    @staticmethod
    def brightness_contrast(sep, val):
        defaults = {'b': 0., 'c': 0.}
        parsed = parse_parameter_list(val, ['b', 'c'], [float, float],
                                      num_required=1, mode='+')
        return defaults | parsed

    @staticmethod
    def clahe(sep, val):
        defaults = {'kernel_size': None, 'nbins': 256}
        parsed = parse_parameter_list(val, ['kernel_size', 'nbins'], [int, int])
        return defaults | parsed

    @staticmethod
    def clip(sep, val):
        defaults = {'a_min': 0., 'a_max': 1.}
        parsed = parse_parameter_list(val, ['a_min', 'a_max'], [float, float])
        return defaults | parsed

    @staticmethod
    def curves(sep, val):
        if sep != '' or val != '':
            raise ValueError('-curves does not take parameters')
        # TODO: currently a hardcoded enhancement curve
        return {'x': [[0., 0.], [0.207, 0.118], [0.513, 0.473], [1., 1.]]}

    @staticmethod
    def exposure(sep, val):
        defaults = {'val': 1.}
        parsed = parse_parameter_list(val, ['val'], [float])
        return defaults | parsed

    @staticmethod
    def gamma(sep, val):
        defaults = {'gamma': 1., 'gain': 1.}
        parsed = parse_parameter_list(val, ['gamma', 'gain'],
                                      [float, float])
        return defaults | parsed

    @staticmethod
    def normalize(sep, val):
        if sep != '' or val != '':
            raise ValueError('-normalize does not take parameters')
        return {}

    @staticmethod
    def shadows(sep, val):
        return parse_parameter_list(val, ['val'], [float], num_required=1,
                                    mode='+')

    @staticmethod
    def sharpen(sep, val):
        defaults = {'radius': 1., 'amount': 1.}
        parsed = parse_parameter_list(val, ['radius', 'amount'], [float, float])
        return defaults | parsed

    @staticmethod
    def stretch(sep, val):
        return parse_parameter_list(val, ['a_min', 'a_max'], [float, float],
                                    num_required=2, mode='+')

    @staticmethod
    def pstretch(sep, val):
        return parse_parameter_list(val, ['min_perc', 'max_perc'],
                                    [float, float], num_required=2, mode='+')


def build_pipeline(cmd_list: CommandList) -> Pipeline:
    """Convert parsed commands into an enhancement pipeline function.

    :param cmd_list: Parsed command list.
    :return: Pipeline callable function.
    """
    pipeline = []
    for (cmd, kwargs) in cmd_list:
        fn = partial(_fn_map[cmd], **kwargs)
        pipeline.append(fn)

    def pipeline_fn(img):
        for p in pipeline:
            img = p(img)
        return img

    return pipeline_fn


def parse_and_build(cmd_strs: CommandStrList) -> Tuple[Pipeline, CommandList]:
    """Parse command arguments and build the pipeline function.

    :param cmd_strs: List of commands returned by argparse.
    :return: Pipeline callable function, parsed command list.
    """
    cmds = parse(cmd_strs, Parsers)
    apply_fn = build_pipeline(cmds)
    return apply_fn, cmds


def add_parser_enhancement_group(parser):
    """Adds the enhancement commands group and positional arguments to the
    given parser."""
    parser.formatter_class = argparse.RawDescriptionHelpFormatter
    opts = parser.add_argument_group('enhancement commands')
    opts.add_argument('commands', metavar='CMD', nargs='*')
    opts.description = commands_description()
    return opts



import numpy as np


def brightness(image, val):
    """Adjust image brightness.

    :param image: Input image.
    :param val: Brightness adjustment factor in units of the image's
           dynamic range (+/-).
    :return: Brightness-adjusted image.
    """
    return image + val


def contrast(image, val: float):
    """Adjust image contrast.

    Note that this is a simple contrast adjustment that corresponds to the
    Legacy method in Photoshop.

    :param image: Input image.
    :param val: Contrast adjustment factor (+/-). Positive values increase
           contrast while negative values decrease contrast. Values below -1
           will invert the image.
    :return: Contrast-adjusted image.
    """
    val = val + 1.
    return val * image + 0.5 * (1. - val)


def brightness_contrast(image: np.ndarray, b: float, c):
    """Apply a brightness-contrast adjustment.

    :param image: Input image.
    :param b: Brightness adjustment factor in units of the image's dynamic
           range (+/-).
    :param c: Contrast adjustment factor (+/-). Positives values increase
           contrast while negative values decrease contrast.
    :return: Brightness/contrast-adjusted image.
    """
    return contrast(brightness(image, b), c)


def exposure(image, val):
    """Increase image exposure.

    :param image: Input image.
    :param val: Exposure adjustment factor (+/-).
    :return: Exposure-adjusted image.
    """
    return image * 2 ** val


def shadows(image, val):
    """Image shadow adjustment.

    Adapted from an implementation by
    `HViktorTsoi <https://gist.github.com/HViktorTsoi/8e8b0468a9fb07842669aa368382a7df>`_.

    :param image: Input image.
    :param val: Shadow adjustment factor (+/-).
    :return: Exposure-adjusted image.
    """
    shadow_val = 1. + val / 100. * 2
    shadow_mid = 3. / 10.
    shadow_region = np.clip(1. - image / shadow_mid, 0, 1)
    shadow_region[np.where(image >= shadow_mid)] = 0
    return (1 - shadow_region) * image + shadow_region * (
            1 - np.power(1 - image, shadow_val))



from skimage.filters import unsharp_mask

sharpen = unsharp_mask
"""Image sharpen using unsharp masking.

See: `skimage.filters.unsharp_mask() <https://scikit-image.org/docs/stable/api/skimage.filters.html#skimage.filters.unsharp_mask>`_
"""


import re
from dataclasses import dataclass

_SHAPE_REGEX = r"(?P<w>\d+)x(?P<h>\d+)"
_ORIGIN_REGEX = r"\+(?P<x>\d+)\+(?P<y>\d+)"
_ROI_REGEX = _SHAPE_REGEX + _ORIGIN_REGEX
_SHAPE_REGEX = re.compile(_SHAPE_REGEX)
_ORIGIN_REGEX = re.compile(_ORIGIN_REGEX)
_ROI_REGEX = re.compile(_ROI_REGEX)


def parse_roi_params(roi_str: str):
    """ Parse an image region-of-interest string of the form :code:`WxH+X+Y`.

    :param roi_str: ROI string
    :return: ROI dataclass
    """
    # ROI return value
    @dataclass
    class ROI:
        x: int = None
        y: int = None
        w: int = None
        h: int = None

        def __str__(self):
            return f'(x:{self.x}, y:{self.y}, w:{self.w}, h:{self.h})'

    # Parse the ROI parameters
    match = _ROI_REGEX.match(roi_str)
    if not match:
        raise ValueError(f'cannot parse ROI from str: {roi_str}')

    # Convert to ints
    roi = ROI()
    for key, value in match.groupdict():
        roi.__dict__[key] = int(value)

    # Return ROI commands
    return roi



import numpy as np
from skimage import img_as_float32, img_as_ubyte, img_as_uint


def as_dtype(image, dtype) -> np.ndarray:
    """Convert an image to a specific fundamental dtype. Automatically performs
    dynamic range adjustment.

    :param image: Input image.
    :param dtype: Numpy dtype. Must be one of: :code:`np.uint8`,
           :code:`np.uint16`, :code:`np.float32`.
    :return: Converted image.
    """
    if dtype == np.uint8:
        image = np.clip(image, a_min=-1., a_max=1.)
        return img_as_ubyte(image)
    elif dtype == np.uint16:
        image = np.clip(image, a_min=-1., a_max=1.)
        return img_as_uint(image)
    elif dtype == np.float32:
        return img_as_float32(image)



import numpy as np


def dynamic_range(t):
    """Returns the dynamic range (minimum and maximum supported value) for the
    provided pixel type. If an array is provided, returns the dynamic range of
    the array's dtype.

    :param t: Array or dtype
    :return: min, max
    """
    if isinstance(t, np.ndarray):
        dtype = t.dtype
    elif t in (np.uint8, np.uint16, np.float32, np.float64):
        dtype = t
    else:
        raise ValueError(f'Unsupported input type: {t}')

    if dtype == np.uint8:
        return 0, 255
    elif dtype == np.uint16:
        return 0, 65535
    elif dtype in (np.float32, np.float64):
        return 0., 1.
    else:
        raise ValueError(f'Unsupported dtype: {t}')
